use proc_macro::TokenStream;
use quote::{format_ident, quote};
use syn::parse_macro_input;

use convert_case::{Case, Casing};

type ColumnList = syn::punctuated::Punctuated<syn::TypePath, syn::Token![,]>;
struct MakeIndexParams {
    unique: Option<syn::Token![!]>,
    name: syn::Ident,
    #[allow(dead_code)]
    comma: syn::Token![,],
    columns: ColumnList,
}

impl syn::parse::Parse for MakeIndexParams {
    fn parse(input: syn::parse::ParseStream) -> syn::Result<Self> {
        Ok(Self {
            unique: input.parse()?,
            name: input.parse()?,
            comma: input.parse()?,
            columns: ColumnList::parse_separated_nonempty(input)?,
        })
    }
}

pub(crate) fn do_make_index(
    tokens: TokenStream,
    microrm_ref: proc_macro2::TokenStream,
) -> TokenStream {
    let input = parse_macro_input!(tokens as MakeIndexParams);

    let index_struct_name = input.name;

    let first_col = input.columns.first().unwrap();
    let mut column_type_path = first_col.path.clone();

    // remove variant name
    column_type_path.segments.pop();
    let last = column_type_path
        .segments
        .pop()
        .expect("Full path to EntityColumn variant");
    column_type_path.segments.push(last.value().clone());

    let index_entity_type_name = format_ident!("{}Entity", index_struct_name);
    let columns = input.columns.clone().into_iter();

    let index_sql_name = format!("{}", index_struct_name).to_case(Case::Snake);

    let unique = input.unique.is_some();

    quote!{
        pub struct #index_struct_name {}
        type #index_entity_type_name = <#column_type_path as #microrm_ref::entity::EntityColumns>::Entity;

        impl #microrm_ref::entity::Index for #index_struct_name {
            type IndexedEntity = #index_entity_type_name;
            fn index_name() -> &'static str {
                #index_sql_name
            }
            fn columns() -> &'static [#column_type_path] where Self: Sized {
                &[#(#columns),*]
            }
            fn unique() -> bool where Self: Sized {
                #unique
            }
        }
    }.into()
}
